import typing
from enum import Enum

import wx


class ViewType(Enum):
    GROUPS = 2
    TREE = 3


class ViewSelect:

    def __init__(self, frame: wx.Frame, groups_item: wx.MenuItem, tree_item: wx.MenuItem, dark_mode: wx.MenuItem,
                 change_callback: typing.Callable):
        self.__groups_item = groups_item
        self.__tree_item = tree_item
        self.__dark_mode = dark_mode
        self.__change_callback = change_callback
        frame.Bind(wx.EVT_MENU, lambda _: self.event(ViewType.GROUPS.value), groups_item)
        frame.Bind(wx.EVT_MENU, lambda _: self.event(ViewType.TREE.value), tree_item)
        frame.Bind(wx.EVT_MENU, self.event_dark_mode, dark_mode)

        self.__groups_item.Check(False)
        self.__tree_item.Check(True)

    def event(self, selected: int):
        self.selected = selected
        self.__change_callback()

    def event_dark_mode(self, _):
        self.__change_callback()

    @property
    def selected(self) -> ViewType:
        return ViewType.GROUPS if self.__groups_item.IsChecked() else ViewType.TREE

    @selected.setter
    def selected(self, selected: int):
        self.__groups_item.Check(selected == ViewType.GROUPS.value)
        self.__tree_item.Check(selected == ViewType.TREE.value)

    @property
    def dark_mode(self) -> bool:
        return self.__dark_mode.IsChecked()

    @dark_mode.setter
    def dark_mode(self, dark_mode: bool):
        self.__dark_mode.Check(dark_mode)



from typing import Optional, List, Callable

import wx

from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.view.AbstractView import ContainerView, ListView, ChoiceView
from asn1editor.wxPython.WxPythonViews import WxPythonView, ControlList


class WxPythonContainerView(WxPythonView, ContainerView):
    icon = 'sequence'

    def __init__(self, name: str, controls: ControlList, parent: wx.Window):
        super(WxPythonContainerView, self).__init__(name, controls, container=True)
        self._children: List[WxPythonView] = []
        self._parent = parent

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        if recursive and self.get_has_value() and self._controls['name'].IsShown():
            sizer = wx.StaticBoxSizer(wx.VERTICAL, self._parent, self._name)
        else:
            sizer = wx.BoxSizer(wx.VERTICAL)

        if 'optional' not in self._controls:
            self._controls['icon'].Show(False)
            self._controls['name'].Show(False)
        else:
            name_sizer = wx.BoxSizer(wx.HORIZONTAL)
            name_sizer.Add(self._controls['icon'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=2)
            name_sizer.Add(self._controls['name'], flag=wx.ALL, border=5)
            sizer.Add(name_sizer)

        if recursive:
            columns = 2 if len(self._children) >= 4 else 1
            container_sizer = wx.FlexGridSizer(cols=columns, vgap=8, hgap=8)
            sizer.Add(container_sizer)
        else:
            container_sizer = sizer

        for child in self._children:
            if recursive or not child._container:
                container_sizer.Add(child.get_sizer(recursive))

        return sizer

    def add_child(self, view: WxPythonView):
        self._children.append(view)
        view.set_visible(self.get_has_value(), recursive=True)

    def enable(self, enabled: bool):
        for child in self._children:
            if child.get_has_value():
                child.enable(enabled)
            child.set_visible(enabled, recursive=True)
        self.structure_changed()

    def get_children(self) -> List[WxPythonView]:
        return self._children

    def destroy(self):
        super(WxPythonContainerView, self).destroy()
        for child in self._children:
            child.destroy()

    def set_visible(self, visible, recursive=True):
        super(WxPythonContainerView, self).set_visible(visible, recursive)
        for child in self._children:
            if recursive or not child._container:
                child.set_visible(visible and self.get_has_value(), recursive)


class WxPythonListView(WxPythonContainerView, ListView, ValueInterface):
    icon = 'sequence_of'

    def __init__(self, name: str, controls: ControlList, parent: wx.Window):
        super(WxPythonListView, self).__init__(name, controls, parent)

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self._controls['value'].GetTopLevelParent().Freeze()
            callback()
            self._controls['value'].GetTopLevelParent().Thaw()

        self._controls['value'].Bind(wx.EVT_SPINCTRL, event_closure)

    def get_value(self) -> str:
        return self._controls['value'].GetValue()

    def set_value(self, val: str):
        self._controls['value'].SetValue(val)

    def enable(self, enabled: bool):
        self._controls['value'].Enable(enabled)
        super(WxPythonListView, self).enable(enabled)

    def add(self, view: WxPythonView):
        self._children.append(view)
        view.set_visible(self.get_has_value(), recursive=True)
        self.structure_changed()

    def remove(self, view: WxPythonView):
        self._children.remove(view)
        view.destroy()
        self.structure_changed()

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        sizer = self._create_sizer(wx.VERTICAL)
        sub_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sub_sizer.Add(self._controls['num_elements'], border=5, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        sub_sizer.Add(self._controls['value'], border=5, flag=wx.ALL)
        sizer.Add(sub_sizer)

        if recursive and self.get_has_value() and self._controls['name'].IsShown():
            content = wx.StaticBoxSizer(wx.VERTICAL, self._parent, self._name)
        else:
            content = wx.BoxSizer(wx.VERTICAL)
        for child in self._children:
            if recursive or not child._container:
                child_sizer = child.get_sizer(recursive)
                content.Add(child_sizer)
        sizer.Add(content)

        return sizer


class WxPythonChoiceView(WxPythonView, ChoiceView, ValueInterface):
    icon = 'choice'

    def __init__(self, name: str, controls: ControlList):
        super(WxPythonChoiceView, self).__init__(name, controls, True)
        self._view: Optional[WxPythonView] = None

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self._controls['value'].GetTopLevelParent().Freeze()
            callback()
            self._controls['value'].GetTopLevelParent().Thaw()

        self._controls['value'].Bind(wx.EVT_COMBOBOX, event_closure)

    def get_value(self) -> str:
        return self._controls['value'].GetStringSelection()

    def set_value(self, val: str):
        self._controls['value'].SetStringSelection(val)

    def enable(self, enabled: bool):
        self._controls['value'].Enable(enabled)
        if self._view is not None:
            self._view.set_visible(enabled)
            self.structure_changed()

    def set_view(self, view: WxPythonView):
        if self._view is not None:
            self._view.destroy()

        self._view = view

        self._view.set_visible(self.get_has_value())

        self.structure_changed()

    def get_view(self) -> WxPythonView:
        return self._view

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        outer_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer = self._create_sizer(wx.VERTICAL)
        sizer.Add(self._controls['value'], border=5)
        outer_sizer.Add(sizer)
        if recursive or not self._view._container:
            content_sizer = wx.BoxSizer(wx.VERTICAL)
            view_sizer = self._view.get_sizer(recursive)
            content_sizer.Add(view_sizer)
            if not self.get_has_value():
                content_sizer.Hide(view_sizer)
            outer_sizer.Add(content_sizer)

        return outer_sizer

    def destroy(self):
        super(WxPythonChoiceView, self).destroy()
        if self._view is not None:
            self._view.destroy()

    def set_visible(self, visible, recursive=True):
        super(WxPythonChoiceView, self).set_visible(visible, recursive)
        if recursive or not self._view._container:
            self._view.set_visible(visible and self.get_has_value(), recursive)



from unittest import TestCase

from asn1editor.controller import Controller, Converter
from tests.controller.test_valueBoolControllers import TestValueInterface, TestOptionalInterface


class TestChoiceInstanceFactory:
    def __init__(self):
        self.value = None
        self.instance = None
        self.member = None

    def create(self, member: str, p: Controller.ListController):
        self.member = member
        self.value = TestValueInterface()
        self.instance = Controller.ValueController(member, p, self.value, None, Converter.Str(0, None))


class TestListController(TestCase):
    def test_init(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()

        Controller.ChoiceController('test', root, value_interface, None, TestChoiceInstanceFactory(), 'choice')
        self.assertEqual(value_interface.val, 'choice')

    def test_add_controller(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()

        controller = Controller.ChoiceController('test', root, value_interface, None, TestChoiceInstanceFactory(), 'choice')
        controller.add_controller('sub', controller)
        self.assertEqual(controller._controller, controller)

    def test_model_to_view(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()
        choice_instance_factory = TestChoiceInstanceFactory()

        controller = Controller.ChoiceController('test', root, value_interface, None, choice_instance_factory, 'choice')

        with self.assertRaises(AssertionError):
            controller.model_to_view({})

        controller = Controller.ChoiceController('test', root, value_interface, optional_interface, choice_instance_factory, 'choice')

        controller.model_to_view({})
        self.assertFalse(optional_interface.val)

        controller.model_to_view({'test': ('choice2', '2')})
        self.assertEqual(value_interface.val, 'choice2')
        self.assertEqual(choice_instance_factory.member, 'choice2')
        self.assertTrue(optional_interface.val)
        self.assertEqual(choice_instance_factory.value.val, '2')

    def test_view_to_model(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()
        choice_instance_factory = TestChoiceInstanceFactory()

        controller = Controller.ChoiceController('test', root, value_interface, optional_interface, choice_instance_factory, 'choice')

        optional_interface.val = False
        self.assertIsNone(controller.view_to_model())

        optional_interface.val = True
        choice_instance_factory.value.val = '1'
        self.assertEqual(controller.view_to_model(), ('choice', '1'))

        optional_interface.val = True
        value_interface.val = 'choice2'
        choice_instance_factory.value.val = '2'
        self.assertEqual(controller.view_to_model(), ('choice2', '2'))

    def test_event_handlers(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()
        choice_instance_factory = TestChoiceInstanceFactory()

        controller = Controller.ChoiceController('test', root, value_interface, optional_interface, choice_instance_factory, 'choice')

        value_interface.val = 'choice2'
        controller.event_handler()
        self.assertEqual(choice_instance_factory.member, 'choice2')



import glob
import json
import locale
import os
import re
import xml.dom.minidom
from typing import List, Tuple, Dict, Union

import asn1tools

from asn1editor.ViewControllerFactory import ViewControllerFactory
from asn1editor.controller.Controller import Controller
from asn1editor.view.AbstractView import AbstractView
from asn1editor.view.AbstractViewFactory import AbstractViewFactory


class ASN1SpecHandler:
    IMPORTS_REGEX = re.compile(r'IMPORTS\s*[\S,\s]*\s*FROM\s*(\S*);', flags=re.MULTILINE)

    def __init__(self, file_name: Union[str, List[str]]):
        # This is necessary to enable parsing of stored dates
        try:
            locale.setlocale(locale.LC_TIME, 'C')
        except locale.Error:
            pass

        if isinstance(file_name, str):
            import_names = []
            # Pre process for import statements to automatically resolve other files
            my_path = os.path.split(os.path.abspath(file_name))[0]
            with open(file_name, 'r') as f:
                content = f.read()
                matches = re.finditer(self.IMPORTS_REGEX, content)
                for match in matches:
                    import_type = match.group(1).lower()
                    dir_files = glob.glob(os.path.join(my_path, '*.asn'))
                    for dir_file in dir_files:
                        dir_file_name = os.path.splitext(os.path.basename(dir_file))[0].lower()
                        if import_type in dir_file_name or dir_file_name in import_type:
                            import_names.append(dir_file)
            self.__file_name = [file_name] + import_names
        else:
            self.__file_name = file_name
        self.__compiled = {}
        self.__type_name = None

    def get_filename(self) -> List[str]:
        return self.__file_name

    def get_types(self, ) -> List[str]:
        types = []
        compiled = self.get_compiled('oer')
        for module_name, module in compiled.modules.items():
            for type_name, compiled_type in module.items():
                types.append(module_name + '.' + type_name)
        return sorted(types)

    def get_compiled(self, codec: str) -> asn1tools.compiler.Specification:
        if codec not in self.__compiled:
            self.__compiled[codec] = asn1tools.compile_files(self.__file_name, codec)
        return self.__compiled[codec]

    def create_mvc_for_type(self, load_type: str, view_factory: AbstractViewFactory) -> Tuple[AbstractView, Controller]:
        compiled = self.get_compiled('oer')
        for module_name, module in compiled.modules.items():
            for type_name, compiled_type in module.items():

                if module_name + '.' + type_name == load_type:
                    mvc_factory = ViewControllerFactory(view_factory)
                    self.__type_name = type_name
                    return mvc_factory.create(compiled_type)

        raise Exception(f'Requested type {load_type} not found in ASN.1 spec')

    @staticmethod
    def get_extensions() -> List[str]:
        return ['*.json', '*.jer', '*.oer', '*.xml', '*.xer', '*.der', '*.ber', '*.per', '*.uper']

    @staticmethod
    def __get_codec(file_name: str) -> str:
        extension_to_codec = {'.json': 'jer', '.jer': 'jer', '.oer': 'oer', '.xer': 'xer', '.xml': 'xer', '.der': 'der', '.ber': 'ber', '.per': 'per',
                              '.uper': 'uper'}
        extension = os.path.splitext(file_name)[1]
        if extension not in extension_to_codec:
            raise Exception(f'Unknown extension {extension}: No ASN.1 codec found')
        return extension_to_codec[extension]

    def load_data_file(self, file_name: str) -> Dict:
        with open(file_name, 'rb') as f:
            return self.get_model_from_data(f.read(), self.__get_codec(file_name))

    def save_data_file(self, file_name: str, model: Dict):
        codec = self.__get_codec(file_name)
        data = self.get_data_from_model(model, codec)
        # Pretty printing of JSON or XML files
        if codec == 'jer':
            data = json.dumps(json.loads(data), indent=4).encode()
        if codec == 'xer':
            dom = xml.dom.minidom.parseString(data)
            data = dom.toprettyxml().encode()
        with open(file_name, 'wb') as f:
            f.write(data)

    def get_data_from_model(self, model: Dict, codec: str) -> bytes:
        compiled = self.get_compiled(codec)
        return compiled.encode(self.__type_name, model[self.__type_name], check_constraints=True)

    def get_model_from_data(self, data: bytes, codec: str) -> Dict:
        compiled = self.get_compiled(codec)
        return {self.__type_name: compiled.decode(self.__type_name, data)}



import datetime
from typing import Callable

import wx

from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.wxPython.WxPythonViews import WxPythonValueView, ControlList, WxPythonView


class WxPythonDateView(WxPythonValueView):
    def __init__(self, name: str, controls: ControlList):
        super(WxPythonDateView, self).__init__(name, controls)

    def get_value(self) -> datetime.date:
        dt: wx.DateTime = self._controls['value'].GetValue()
        return datetime.date(year=dt.GetYear(), month=dt.GetMonth(), day=dt.GetDay())

    def set_value(self, val: datetime.date):
        self._controls['value'].SetValue(wx.DateTime(day=val.day, month=val.month, year=val.year))


class WxPythonTimeView(WxPythonValueView):
    def __init__(self, name: str, controls: ControlList):
        super(WxPythonTimeView, self).__init__(name, controls)

    def get_value(self) -> datetime.time:
        hour, minute, second = self._controls['value'].GetTime()
        return datetime.time(hour=hour, minute=minute, second=second)

    def set_value(self, val: datetime.time):
        self._controls['value'].SetTime(val.hour, val.minute, val.second)


class WxPythonDateTimeView(WxPythonView, ValueInterface):
    def __init__(self, name: str, controls: ControlList):
        super(WxPythonDateTimeView, self).__init__(name, controls)

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        sizer = self._create_sizer()
        sizer.Add(self._controls['value'], proportion=1, flag=wx.ALL | wx.EXPAND, border=5)
        sizer.Add(self._controls['time'], proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        if self._controls.get('style') == 'hidden':
            sizer.ShowItems(False)
        return sizer

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()

        self._controls['value'].Bind(wx.EVT_TEXT, event_closure)
        self._controls['time'].Bind(wx.EVT_TEXT, event_closure)

    def get_value(self) -> datetime.datetime:
        dt: wx.DateTime = self._controls['value'].GetValue()
        hour, minute, second = self._controls['time'].GetTime()
        return datetime.datetime(year=dt.GetYear(), month=dt.GetMonth(), day=dt.GetDay(),
                                 hour=hour, minute=minute, second=second)

    def set_value(self, val: datetime.datetime):
        self._controls['value'].SetValue(wx.DateTime(day=val.day, month=val.month, year=val.year))
        self._controls['time'].SetTime(val.hour, val.minute, val.second)



import typing

import wx

from asn1editor.wxPython import Resources
from asn1editor.wxPython.WxPythonComplexViews import WxPythonContainerView, WxPythonChoiceView
from asn1editor.wxPython.WxPythonViews import WxPythonView


class TreeView:

    def __init__(self, window: wx.Window, content_window: wx.ScrolledWindow, root_name: str):
        self.__tree_ctrl = wx.TreeCtrl(window)
        Resources.get_bitmap_from_svg('root')
        root_item = self.__tree_ctrl.AddRoot(root_name, Resources.image_list.get_index('root'))
        self.__tree_ctrl.SetItemBold(root_item, True)
        self.__tree_ctrl.Bind(wx.EVT_TREE_SEL_CHANGED, self.item_selected)
        self.__tree_ctrl.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, self.item_right_clicked)
        self.__tree_ctrl.SetImageList(Resources.image_list.get_image_list())
        self.__content_window = content_window
        self.__current_view: typing.Optional[WxPythonView] = None

    def __sync(self, tree_item: wx.TreeItemId, view: WxPythonView):
        if isinstance(view, WxPythonContainerView):
            # First, check if the view is not in the tree yet
            container_item_for_view = self.__add_if_not_in_tree(tree_item, view)

            # Now check if it was removed from the tree
            self.__delete_if_removed(container_item_for_view, view.get_children())

            # Finally handle children
            if view.get_has_value():
                for child in view.get_children():
                    self.__sync(container_item_for_view, child)
        if isinstance(view, WxPythonChoiceView):
            container_item_for_view = self.__add_if_not_in_tree(tree_item, view)

            self.__delete_if_removed(container_item_for_view, [view.get_view()])
            if view.get_has_value():
                self.__sync(container_item_for_view, view.get_view())

    def __add_if_not_in_tree(self, tree_item: wx.TreeItemId, view: typing.Union[WxPythonContainerView, WxPythonChoiceView]) -> typing.Optional[wx.TreeItemId]:
        # First, check if the view is not in the tree yet
        found = False
        container_item_for_view = None
        tree_child, cookie = self.__tree_ctrl.GetFirstChild(tree_item)
        while tree_child.IsOk():
            child_view = self.__tree_ctrl.GetItemData(tree_child)
            if child_view == view:
                container_item_for_view = tree_child
                found = True
            tree_child, cookie = self.__tree_ctrl.GetNextChild(tree_child, cookie)
        if not found:
            image = Resources.image_list.get_index(view.icon)
            container_item_for_view = self.__tree_ctrl.AppendItem(tree_item, view.get_name(), image=image)
            self.__tree_ctrl.SetItemData(container_item_for_view, view)

        self.__tree_ctrl.SetItemBold(container_item_for_view, view.get_has_value())

        return container_item_for_view

    def __delete_if_removed(self, container_item_for_view: wx.TreeItemId, views: typing.List[WxPythonView]):
        current_child_views_in_tree = []
        tree_child, cookie = self.__tree_ctrl.GetFirstChild(container_item_for_view)
        while tree_child.IsOk():
            current_child_views_in_tree.append((self.__tree_ctrl.GetItemData(tree_child), tree_child))
            tree_child, cookie = self.__tree_ctrl.GetNextChild(tree_child, cookie)

        for current_child, current_treeitem in current_child_views_in_tree:
            if current_child not in views:
                self.__tree_ctrl.Delete(current_treeitem)

    def get_ctrl(self, root_view: WxPythonView) -> wx.TreeCtrl:
        self.__sync(self.__tree_ctrl.GetRootItem(), root_view)
        root_view.set_visible(False, recursive=True)

        selected = self.__tree_ctrl.GetSelection()
        if selected.IsOk() and self.__tree_ctrl.GetItemData(selected) is not None:
            self.__show_view(self.__tree_ctrl.GetItemData(selected))

        self.__tree_ctrl.Show()
        return self.__tree_ctrl

    def hide(self):
        self.__current_view = None
        self.__tree_ctrl.Hide()

    def destroy(self):
        self.__tree_ctrl.Destroy()

    def item_selected(self, e: wx.TreeEvent):
        view = self.__tree_ctrl.GetItemData(e.GetItem())
        self.__show_view(view)

    def item_right_clicked(self, e: wx.TreeEvent):
        class RightClickMenu(wx.Menu):
            def __init__(self, parent):
                super(RightClickMenu, self).__init__()
                self.parent = parent

                self.expand = self.Append(wx.NewId(), 'Expand')
                self.expand_all = self.Append(wx.NewId(), 'Expand all')
                self.AppendSeparator()
                self.collapse = self.Append(wx.NewId(), 'Collapse')
                self.collapse_all = self.Append(wx.NewId(), 'Collapse all')

        menu = RightClickMenu(self.__tree_ctrl.GetTopLevelParent())
        menu.Bind(wx.EVT_MENU, lambda _: self.__tree_ctrl.ExpandAllChildren(e.GetItem()), menu.expand)
        menu.Bind(wx.EVT_MENU, lambda _: self.__tree_ctrl.ExpandAll(), menu.expand_all)
        menu.Bind(wx.EVT_MENU, lambda _: self.__tree_ctrl.CollapseAllChildren(e.GetItem()), menu.collapse)
        menu.Bind(wx.EVT_MENU, lambda _: self.__tree_ctrl.CollapseAll(), menu.collapse_all)
        self.__tree_ctrl.GetTopLevelParent().PopupMenu(menu, e.GetPoint())

    def __show_view(self, view: typing.Optional[WxPythonView]):
        self.__content_window.GetTopLevelParent().Freeze()

        if self.__current_view is not None:
            self.__current_view.set_visible(False, recursive=False)

        self.__current_view = view
        if view is not None:
            view.set_visible(True, recursive=False)

        sizer: wx.Sizer = self.__content_window.GetSizer()
        sizer.Clear()

        if view is not None:
            sizer.Add(view.realize().get_sizer(recursive=False), 0, wx.ALL | wx.EXPAND, 5)

        sizer.Layout()

        self.__content_window.SetSizer(sizer)
        self.__content_window.FitInside()
        self.__content_window.AdjustScrollbars()

        self.__content_window.GetTopLevelParent().Thaw()



from typing import List, Tuple, Optional, Union

import wx
import wx.adv
import wx.lib.masked.numctrl
import wx.svg

from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.view.AbstractView import AbstractView, ContainerView, ListView, ChoiceView
from asn1editor.view.AbstractViewFactory import AbstractViewFactory
from asn1editor.wxPython import Resources
from asn1editor.wxPython.Styler import Styler
from asn1editor.wxPython.WxPythonComplexViews import WxPythonContainerView, WxPythonListView, WxPythonChoiceView
from asn1editor.wxPython.WxPythonDateTimeViews import WxPythonDateView, WxPythonTimeView, WxPythonDateTimeView
from asn1editor.wxPython.WxPythonViews import WxPythonValueView, WxPythonBooleanView, \
    WxPythonBitstringView, WxPythonHexStringView, WxPythonValueSelectionView, ControlList


class WxPythonViewFactory(AbstractViewFactory):
    def __init__(self, window: wx.ScrolledWindow, styler: Styler):
        self._window = window
        self._styler = styler

    def get_enumerated_view(self, name: str, choices: List[str], optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, ':', 'enumerated')

        controls['value'] = wx.ComboBox(self._window, choices=choices, style=wx.CB_READONLY)
        self._apply_style(controls)

        view = WxPythonValueSelectionView(name, controls)
        return view, view, view if optional else None

    def get_text_view(self, name: str, text: str) -> AbstractView:
        controls = self._get_controls(name, False)

        controls['value'] = wx.StaticText(self._window, wx.ID_ANY, text)
        self._apply_style(controls)

        view = WxPythonValueView(name, controls)

        return view

    def get_container_view(self, name: str, optional: bool) -> Tuple[ContainerView, OptionalInterface]:
        controls = self._get_controls(name, optional, icon=WxPythonContainerView.icon)

        view = WxPythonContainerView(name, controls, self._window)

        return view, view if optional else None

    def get_list_view(self, name: str, minimum: int, maximum: int, optional: bool) -> Tuple[ListView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, icon=WxPythonListView.icon)

        num_elements = wx.SpinCtrl(self._window)
        if minimum is not None:
            num_elements.SetMin(minimum)
        else:
            minimum = 0
        if maximum is not None:
            num_elements.SetMax(maximum)
        else:
            maximum = 'infinite'
        num_elements.SetToolTip(f"Minimum elements: {minimum}, maximum elements: {maximum}")
        controls['value'] = num_elements
        controls['num_elements'] = wx.StaticText(self._window, wx.ID_ANY, "Elements:")
        self._apply_style(controls)

        view = WxPythonListView(name, controls, self._window)

        return view, view, view if optional else None

    def get_number_view(self, name: str, optional: bool, minimum: Optional[Union[int, float]],
                        maximum: Optional[Union[int, float]], float_: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, ':', 'float' if float_ else 'integer')

        edit = wx.lib.masked.numctrl.NumCtrl(self._window)
        tool_tip = []
        if isinstance(minimum, int) or isinstance(minimum, float):
            edit.SetAllowNegative(minimum < 0)
            edit.SetMin(minimum)
            tool_tip.append(f'Minimum: {minimum}')
        if isinstance(maximum, int) or isinstance(maximum, float):
            edit.SetMax(maximum)
            tool_tip.append(f'Maximum: {maximum}')
        if float_:
            edit.SetFractionWidth(6)
        if len(tool_tip):
            edit.SetToolTip(', '.join(tool_tip))

        controls['value'] = edit
        self._apply_style(controls)

        view = WxPythonValueView(name, controls)
        return view, view, view if optional else None

    def get_boolean_view(self, name: str, optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, ':', 'bool')

        controls['value'] = wx.CheckBox(self._window)
        self._apply_style(controls)

        view = WxPythonBooleanView(name, controls)
        return view, view, view if optional else None

    def get_string_view(self, name: str, string_type: str, optional: bool, minimum: Optional[int], maximum: Optional[int]):
        controls = self._get_controls(name, optional, ':', 'string', string_type)

        edit = wx.TextCtrl(self._window)
        if maximum:
            edit.SetMaxLength(maximum)
        else:
            maximum = 'infinite'
        if minimum is None:
            minimum = '0'
        edit.SetToolTip(f"Minimum characters: {minimum}, maximum characters: {maximum}")

        controls['value'] = edit
        self._apply_style(controls)

        view = WxPythonValueView(name, controls)
        return view, view, view if optional else None

    def get_hex_string_view(self, name: str, optional: bool, minimum: Optional[int], maximum: Optional[int]):
        controls = self._get_controls(name, optional, ':', 'string', 'OCTET STRING')

        controls['selector'] = wx.CheckBox(self._window, label='Hex')
        controls['value'] = wx.TextCtrl(self._window)
        self._apply_style(controls)

        view = WxPythonHexStringView(name, controls, minimum, maximum)
        return view, view, view if optional else None

    def get_choice_view(self, name: str, choices: List[str], optional: bool) -> Tuple[ChoiceView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, icon=WxPythonChoiceView.icon)

        controls['value'] = wx.ComboBox(self._window, choices=choices, style=wx.CB_READONLY)
        self._apply_style(controls)

        view = WxPythonChoiceView(name, controls)

        return view, view, view if optional else None

    def get_bitstring_view(self, name: str, number_of_bits: int, named_bits: List[Tuple[str, int]], optional: bool) -> Tuple[AbstractView, BitstringInterface,
                                                                                                                             OptionalInterface]:
        controls = self._get_controls(name, optional, icon='bitstring')

        checkboxes: List[Tuple[int, wx.CheckBox]] = []

        style = self._styler.get_style(name)

        if named_bits:
            for name, bit in named_bits:
                bit_checkbox = wx.CheckBox(self._window, label=f"{bit}: {name}")
                if style == 'read_only':
                    bit_checkbox.Enable(False)
                checkboxes.append((bit, bit_checkbox))
        else:
            for bit in range(number_of_bits):
                bit_checkbox = wx.CheckBox(self._window, label=str(bit))
                if style == 'read_only':
                    bit_checkbox.Enable(False)
                checkboxes.append((bit, bit_checkbox))

        controls['checkboxes'] = checkboxes

        view = WxPythonBitstringView(name, controls, self._window)

        return view, view, view if optional else None

    def get_date_view(self, name: str, optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, ':', 'date')

        controls['value'] = wx.adv.DatePickerCtrl(self._window)
        self._apply_style(controls)

        view = WxPythonDateView(name, controls)
        return view, view, view if optional else None

    def get_time_view(self, name: str, optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, ':', 'date')

        controls['value'] = wx.adv.TimePickerCtrl(self._window)
        self._apply_style(controls)

        view = WxPythonTimeView(name, controls)
        return view, view, view if optional else None

    def get_datetime_view(self, name: str, optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        controls = self._get_controls(name, optional, ':', 'date')

        controls['value'] = wx.adv.DatePickerCtrl(self._window)
        controls['time'] = wx.adv.TimePickerCtrl(self._window)
        self._apply_style(controls)

        view = WxPythonDateTimeView(name, controls)
        return view, view, view if optional else None

    def _get_controls(self, name: str, optional: bool, suffix: str = '', icon: str = None, icon_tooltip: str = None) -> \
            ControlList:
        controls = {}
        if optional:
            control = wx.CheckBox(self._window, wx.ID_ANY, name + suffix)
            control.SetToolTip("Optional element")
            controls['optional'] = control
            controls['name'] = control
        else:
            control = wx.StaticText(self._window, wx.ID_ANY, name + suffix)
            controls['name'] = control
        if icon is not None:
            controls['icon'] = self._get_svg(icon, icon_tooltip)

        style = self._styler.get_style(name)
        if style is not None:
            controls['style'] = style

        return controls

    def _get_svg(self, bitmap_name: str, icon_tooltip: str = None) -> wx.StaticBitmap:
        bitmap = Resources.image_list.get_bitmap(bitmap_name)
        if bitmap is None:
            bitmap = Resources.get_bitmap_from_svg(bitmap_name)

        static_bitmap = wx.StaticBitmap(self._window, bitmap=bitmap)

        if icon_tooltip:
            static_bitmap.SetToolTip(icon_tooltip)
        else:
            static_bitmap.SetToolTip(bitmap_name.upper().replace('_', ' '))
        return static_bitmap

    @staticmethod
    def _apply_style(controls: ControlList):
        if controls.get('style') == 'read_only' and 'value' in controls:
            controls['value'].Enable(False)



import typing

from asn1editor import PluginInterface


class Plugin:  # pragma: no cover
    def get_name(self) -> str:
        """
        Returns the name of a plugin - used to name the menu item folder
        """
        raise NotImplementedError

    def get_menus(self) -> typing.List[typing.Tuple[str, typing.Callable]]:
        """
        Get the menus required by the plugin.

        Returns a list of strings and callables, where the string is the menu caption and
        the callable a callback that will be executed when the menu item is selected.
        """
        raise NotImplementedError

    def get_tools(self) -> typing.List[typing.Tuple[str, str, str, typing.Callable]]:
        """
        Get the toolbar items required by the plugin.

        Returns a list of strings and callables, where the strings are the label, the tool tip
        and a path to an icon, the callable a callback that will be executed when the menu
        item is selected.
        """
        raise NotImplementedError

    def get_about(self) -> typing.Optional[str]:
        """
        Gets s string of the plugin to be added to the about box

        """
        raise NotImplementedError

    def connect(self, plugin_interface: PluginInterface):
        """
        Connects the plugin host and the plugin.

        The function will be called during startup and serves to register the PluginInterface.

        :param plugin_interface: Object that implements the PluginInterface.
        """
        raise NotImplementedError



import datetime
from typing import Optional, Union

from asn1tools.codecs import oer

from asn1editor.controller import Controller, Converter
from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface


class ControllerFactory:
    def __init__(self, parent: Controller):
        self._parent = parent

    def create_value_controller(self, type_: oer.Type, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface],
                                minimum: Optional[Union[str, int, float]] = 0):
        if isinstance(type_, oer.Integer):
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface, Converter.Int(minimum, type_.default))
        elif isinstance(type_, oer.Real):
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface, Converter.Float(minimum, type_.default))
        elif isinstance(type_, oer.Enumerated):
            if type_.default is None:
                default = sorted(type_.value_to_data.values())[0]
            else:
                default = type_.default
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface, Converter.Str(0, default))
        elif type(type_) in [oer.UTF8String, oer.VisibleString, oer.GeneralString, oer.ObjectIdentifier, oer.IA5String]:
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface, Converter.Str(minimum, type_.default))
        elif isinstance(type_, oer.OctetString) or isinstance(type_, oer.BitString):
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface, Converter.ByteString(minimum, type_.default))
        elif isinstance(type_, oer.Date):
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface, Converter.Any(0, datetime.date.today()))
        elif isinstance(type_, oer.TimeOfDay):
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface,
                                                    Converter.Any(0, datetime.datetime.now().time()))
        elif isinstance(type_, oer.DateTime) or isinstance(type_, oer.UTCTime) or isinstance(type_, oer.GeneralizedTime):
            controller = Controller.ValueController(type_.name, self._parent, value_interface, optional_interface,
                                                    Converter.Any(0, datetime.datetime.now()))
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

        self.__register_events(controller, value_interface, optional_interface)

    def create_container_controller(self, type_: oer.Type, optional_interface: Optional[OptionalInterface]) -> Controller.ContainerController:
        if isinstance(type_, oer.Sequence) or isinstance(type_, oer.Set):
            controller = Controller.ContainerController(type_.name, self._parent, optional_interface)
            if optional_interface is not None:
                optional_interface.register_optional_event(controller.optional_handler)

            return controller
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

    def create_list_controller(self, type_: oer.Type, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface], list_instance_factory,
                               minimum_elements: int):
        if isinstance(type_, oer.SequenceOf) or isinstance(type_, oer.SetOf):
            controller = Controller.ListController(type_.name, self._parent, value_interface, optional_interface, list_instance_factory, minimum_elements)
            self.__register_events(controller, value_interface, optional_interface)
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

    def create_bool_controller(self, type_: oer.Type, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface]):
        if isinstance(type_, oer.Boolean):
            controller = Controller.BoolController(type_.name, self._parent, value_interface, optional_interface, type_.default)
            self.__register_events(controller, value_interface, optional_interface)
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

    def create_choice_controller(self, type_: oer.Type, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface],
                                 choice_instance_factory):
        def has_no_recursive_member(m):
            if isinstance(m, oer.Recursive):
                return False
            elif type(m) in [oer.Set, oer.Sequence]:
                return not any(isinstance(m, oer.Recursive) for m in m.root_members)
            elif type(m) in [oer.SequenceOf, oer.SetOf]:
                return not isinstance(m.element_type, oer.Recursive)
            return True

        if isinstance(type_, oer.Choice):
            if type_.default is None:
                # Do not use recursive types as default
                candidates = list(filter(has_no_recursive_member, type_.members))
                default = sorted([member.name for member in candidates])[0]
            else:
                default = type_.default
            controller = Controller.ChoiceController(type_.name, self._parent, value_interface, optional_interface, choice_instance_factory, default)
            self.__register_events(controller, value_interface, optional_interface)
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

    def create_bitstring_controller(self, type_: oer.Type, bitstring_interface: BitstringInterface, optional_interface: Optional[OptionalInterface]):
        if isinstance(type_, oer.BitString):
            controller = Controller.BitstringController(type_.name, self._parent, bitstring_interface, optional_interface, type_.number_of_bits)
            if optional_interface is not None:
                optional_interface.register_optional_event(controller.optional_handler)
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

    def create_null_controller(self, type_: oer.Type):
        if isinstance(type_, oer.Null):
            Controller.NullController(type_.name, self._parent, None)
        else:
            raise Exception(f"Unknown type for ControllerFactory: {type_}")

    @staticmethod
    def __register_events(controller: Controller.Controller, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface]):
        value_interface.register_change_event(controller.event_handler)
        if optional_interface is not None:
            optional_interface.register_optional_event(controller.optional_handler)



__version__ = "0.5.1"
# noinspection SpellCheckingInspection
__author__ = "Florian Fetz"

import argparse

import wx

from .Plugin import Plugin
from .PluginInterface import PluginInterface
from .wxPython import MainWindow as WxPythonMainWindow

__all__ = ['WxPythonMainWindow', 'Plugin', 'PluginInterface']


def _wx_python_editor():
    parser = argparse.ArgumentParser(description='ASN.1 editor')
    parser.add_argument('asn1spec', nargs='?', help='ASN.1 specification file name')
    parser.add_argument('-type', required=False, help='Name of the ASN.1 type to load (Module name.Type name)')
    parser.add_argument('-data', required=False, help='Data file to load')

    args = parser.parse_args()

    app = wx.App()

    frame = WxPythonMainWindow()
    if args.asn1spec is not None:
        frame.load_spec(args.asn1spec, args.type)
    if args.data is not None:
        frame.load_data_from_file(args.data)

    frame.Show()

    app.MainLoop()



import typing


class ValueInterface:  # pragma: no cover
    def register_change_event(self, callback: typing.Callable):
        raise NotImplementedError

    def get_value(self) -> typing.Union[str, bytes, bool]:
        raise NotImplementedError

    def set_value(self, val: typing.Union[str, bytes, bool]):
        raise NotImplementedError



from unittest import TestCase

from asn1editor.controller import Controller

class TestValueController(TestCase):
    def test_controller(self):
        root = Controller.RootController('root')

        controller = Controller.Controller('Test', root, None)

        with self.assertRaises(NotImplementedError):
            controller.add_controller('Error', root)

        with self.assertRaises(NotImplementedError):
            controller.model_to_view({})

        with self.assertRaises(NotImplementedError):
            controller.view_to_model()

        controller.optional_handler()

        self.assertEqual(str(controller), 'Test')

    def test_nullController(self):
        root = Controller.RootController('root')

        controller = Controller.NullController('Test', root, None)

        with self.assertRaises(Exception):
            controller.add_controller('Test', root)



from typing import Optional, Any, Dict, List, Tuple, Union

from asn1editor.controller import Converter
from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface


class Controller:
    def __init__(self, name: str, parent: 'Controller', optional_interface: Optional[OptionalInterface]):
        self._name = name
        self._parent = parent
        self._optional_interface = optional_interface
        if self._optional_interface:
            self._optional_interface.set_has_value(False)
        if parent is not None:
            parent.add_controller(name, self)

    def add_controller(self, name: str, other: 'Controller'):
        raise NotImplementedError()

    def model_to_view(self, model: Dict[str, Any]):
        raise NotImplementedError()

    def view_to_model(self) -> Optional[Dict[str, Any]]:
        raise NotImplementedError()

    def event_handler(self):
        pass

    def optional_handler(self):
        pass

    def _model_to_view_optional(self, model: Dict[str, Any]):
        has_value = self._name in model or isinstance(self._parent, ListController) or isinstance(self._parent, ChoiceController)
        assert has_value or self._optional_interface
        if self._optional_interface:
            # Notify view if value is there or not
            self._optional_interface.set_has_value(has_value)
        return has_value

    def _view_to_model_optional(self):
        return not self._optional_interface or self._optional_interface.get_has_value()

    def __repr__(self):
        return self._name


class ValueController(Controller):
    def __init__(self, name: str, parent: Controller, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface],
                 data_converter: Converter.Converter):
        super().__init__(name, parent, optional_interface)
        self._value_interface = value_interface
        self._data_converter = data_converter
        self._default = self._data_converter.default()
        self._set_value(self._default)

    def add_controller(self, name: str, other: Controller):
        raise Exception('ValueController cannot add a controller')

    def model_to_view(self, model: Union[Dict[str, Any], Any]):
        if isinstance(model, Dict):
            if self._model_to_view_optional(model):
                self._set_value(model[self._name])
        else:
            self._set_value(model)

    def view_to_model(self) -> Any:
        if self._view_to_model_optional():
            return self._data_converter.from_view(self._value_interface.get_value())

    def _set_value(self, value: Any):
        self._value_interface.set_value(self._data_converter.to_view(value))


class BoolController(Controller):
    def __init__(self, name: str, parent: Controller, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface], default: bool):
        super().__init__(name, parent, optional_interface)
        self._value_interface = value_interface
        self._default = default if default is not None else False
        self._value_interface.set_value(self._default)

    def add_controller(self, name: str, other: Controller):
        raise Exception('ValueController cannot add a controller')

    def model_to_view(self, model: Union[Dict[str, Any], Any]):
        if isinstance(model, Dict):
            if self._model_to_view_optional(model):
                self._value_interface.set_value(model[self._name])
        else:
            self._value_interface.set_value(model)

    def view_to_model(self) -> bool:
        if self._view_to_model_optional():
            return self._value_interface.get_value()


class ListController(Controller):
    def __init__(self, name: str, parent: Controller, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface], list_instance_factory,
                 default: int):
        super().__init__(name, parent, optional_interface)
        self._value_interface = value_interface
        self._controllers = []
        self._list_instance_factory = list_instance_factory
        if default is not None and default > 0 and not optional_interface:
            self._value_interface.set_value(str(default))
            self.event_handler()

    def add_controller(self, name: str, other: Controller):
        self._controllers.append(other)

    def model_to_view(self, model: Union[List[Any], Dict[str, Any]]):
        if self._model_to_view_optional(model):
            if isinstance(model, Dict):
                model: List = model[self._name]
            new_num = len(model)
            self.__sync_controllers(new_num)
            for i, controller in enumerate(self._controllers):
                controller.model_to_view(model[i])
            self._value_interface.set_value(str(new_num))
        else:
            self.__sync_controllers(0)

    def view_to_model(self) -> Optional[List]:
        if self._view_to_model_optional():
            model = []
            for controller in self._controllers:
                model.append(controller.view_to_model())
            return model

    def event_handler(self):
        new_num = int(self._value_interface.get_value())
        self.__sync_controllers(new_num)

    def optional_handler(self):
        if not self._optional_interface.get_has_value():
            self.__sync_controllers(0)
        else:
            self.event_handler()

    def __sync_controllers(self, new_num: int):
        if new_num > len(self._controllers):
            prev_num = len(self._controllers)
            for i in range(prev_num, new_num):
                self._list_instance_factory.create(i, self)
        else:
            for i in reversed(range(new_num, len(self._controllers))):
                # First, delete the view
                self._list_instance_factory.destroy(i)
                # And finally destroy the controller
                del self._controllers[i]


class ChoiceController(Controller):
    def __init__(self, name: str, parent: Controller, value_interface: ValueInterface, optional_interface: Optional[OptionalInterface], choice_instance_factory,
                 default: str):
        super().__init__(name, parent, optional_interface)
        self._value_interface = value_interface
        self._controller: Optional[Controller] = None
        self._choice_instance_factory = choice_instance_factory
        self._value_interface.set_value(default)
        self._choice_instance_factory.create(default, self)

    def add_controller(self, name: str, other: 'Controller'):
        self._controller = other

    def model_to_view(self, model: Dict[str, Any]):
        if self._model_to_view_optional(model):
            choice = model[self._name][0]
            self._value_interface.set_value(choice)
            self._choice_instance_factory.create(choice, self)
            self._controller.model_to_view(model[self._name][1])

    def view_to_model(self) -> Optional[Tuple[str, Any]]:
        if self._view_to_model_optional():
            return self._value_interface.get_value(), self._controller.view_to_model()

    def event_handler(self):
        choice = self._value_interface.get_value()
        self._choice_instance_factory.create(choice, self)


class BitstringController(Controller):
    def __init__(self, name: str, parent: Optional[Controller], bitstring_interface: BitstringInterface,
                 optional_interface: Optional[OptionalInterface], number_of_bits: int):
        super().__init__(name, parent, optional_interface)
        self._bitstring_interface = bitstring_interface
        self._number_of_bits = number_of_bits

    def add_controller(self, name: str, other: Controller):
        raise Exception('ValueController cannot add a controller')

    def model_to_view(self, model: Dict[str, Any]):
        if self._model_to_view_optional(model):
            values = []
            bytes_, num_bits = model[self._name]
            assert num_bits == self._number_of_bits
            for bit in range(self._number_of_bits):
                bit_index = bit % 8
                byte_index = bit // 8
                if bytes_[byte_index] & (1 << bit_index):
                    values.append(bit)
            self._bitstring_interface.set_values(values)

    def view_to_model(self) -> Optional[Tuple[bytes, int]]:
        if self._view_to_model_optional():
            bytes_ = bytearray(b'\x00' * -(-self._number_of_bits // 8))
            values = self._bitstring_interface.get_values()
            for bit in values:
                bit_index = bit % 8
                byte_index = bit // 8
                bytes_[byte_index] |= (1 << bit_index)

            return bytes_, self._number_of_bits


class ContainerController(Controller):
    def __init__(self, name: str, parent: Optional[Controller], optional_interface: Optional[OptionalInterface]):
        super().__init__(name, parent, optional_interface)
        self._controllers: Dict[str, Controller] = {}

    def add_controller(self, name: str, other: Controller):
        self._controllers[name] = other

    def model_to_view(self, model: Dict[str, Any]):
        if self._model_to_view_optional(model):
            if isinstance(self._parent, ListController) or isinstance(self._parent, ChoiceController):
                self._model_to_view(model)
            else:
                self._model_to_view(model[self._name])

    def view_to_model(self) -> Optional[Dict[str, Any]]:
        if self._view_to_model_optional():
            return self._view_to_model()

    def _model_to_view(self, model: Dict[str, Any]):
        for name, controller in self._controllers.items():
            controller.model_to_view(model)

    def _view_to_model(self) -> Dict[str, Any]:
        model = {}
        for name, controller in self._controllers.items():
            sub_model = controller.view_to_model()
            if sub_model is not None or isinstance(controller, NullController):
                model[name] = sub_model
        return model


class NullController(Controller):

    def add_controller(self, name: str, other: 'Controller'):
        raise Exception('NullController cannot add a controller')

    def model_to_view(self, model: Dict[str, Any]):
        pass

    def view_to_model(self) -> Optional:
        return None


class RootController(ContainerController):
    def __init__(self, name: str):
        super().__init__(name, None, None)

    def model_to_view(self, model: Dict[str, Any]):
        self._model_to_view(model)

    def view_to_model(self) -> Optional[Dict[str, Any]]:
        return self._view_to_model()



import json
import os
import sys

settings = {}


def _get_dir() -> str:
    if sys.platform == 'win32':
        win_path = os.path.expandvars(r'%APPDATA%/asn1editor')
        os.makedirs(win_path, exist_ok=True)
        return win_path
    elif sys.platform.startswith('linux'):
        home = os.path.expanduser('~/.local/share/asn1editor')
        os.makedirs(home, exist_ok=True)
        return home
    else:
        return ''


def _get_settings_filename() -> str:
    return os.path.join(_get_dir(), 'settings.json')


def log_error(log_message: str):
    try:
        with open(os.path.join(_get_dir(), 'error_log.txt'), 'a+') as f:
            import datetime
            f.write(f'{datetime.datetime.now()}: {log_message}\n\n')
    finally:
        pass


def load():
    global settings
    try:
        with open(_get_settings_filename(), 'r') as f:
            settings = json.load(f)
    except FileNotFoundError:
        settings = {}
    except json.decoder.JSONDecodeError:
        settings = {}


def save():
    global settings
    with open(_get_settings_filename(), 'w+') as f:
        f.write(json.dumps(settings))



from abc import ABC
from typing import Any


class AbstractView:  # pragma: no cover
    def realize(self) -> Any:
        raise NotImplementedError


class ContainerView(AbstractView, ABC):  # pragma: no cover
    def add_child(self, view: AbstractView):
        raise NotImplementedError


class ListView(AbstractView, ABC):  # pragma: no cover
    def add(self, view: AbstractView):
        raise NotImplementedError

    def remove(self, view: AbstractView):
        raise NotImplementedError


class ChoiceView(AbstractView, ABC):  # pragma: no cover
    def set_view(self, view: AbstractView):
        raise NotImplementedError



import os
import sys
import typing

import wx
import wx.svg

from asn1editor.wxPython.ImageList import ImageList

image_list: typing.Optional[ImageList] = None


def resource_path(relative_path: str) -> str:
    """ Get absolute path to resource required for PyInstaller """
    # noinspection SpellCheckingInspection
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__)))
    path = os.path.join(base_path, relative_path)
    if not os.path.exists(path):
        raise FileNotFoundError(f'File {path} does not exist')
    return path


def plugin_resource_path(relative_path: str) -> str:
    """ Get absolute path to resource required for PyInstaller """
    # noinspection SpellCheckingInspection
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__)))
    path = os.path.join(base_path, '../..', relative_path)
    if not os.path.exists(path):
        raise FileNotFoundError(f'File {path} does not exist')
    return path


def get_bitmap_from_svg(bitmap_name) -> wx.Bitmap:
    # noinspection PyArgumentList
    image: wx.svg.SVGimage = wx.svg.SVGimage.CreateFromFile(resource_path(f'icons/{bitmap_name}.svg'))
    bitmap = image.ConvertToBitmap(width=16, height=16)
    image_list.add_bitmap(bitmap, bitmap_name)
    return bitmap



from unittest import TestCase

from asn1editor.controller import Controller, Converter
from tests.controller.test_valueBoolControllers import TestValueInterface, TestOptionalInterface


class TestListInstanceFactory:
    def __init__(self):
        self.values = {}
        self.instances = {}

    def create(self, i, p: Controller.ListController):
        self.values[i] = TestValueInterface()
        self.instances[i] = Controller.ValueController(str(i), p, self.values[i], None, Converter.Str(0, None))

    def destroy(self, i):
        del self.values[i]
        del self.instances[i]


class TestListController(TestCase):
    def test_init(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()

        Controller.ListController('test', root, value_interface, None, TestListInstanceFactory(), 1)
        self.assertEqual(value_interface.val, '1')

        value_interface.val = 42
        Controller.ListController('test', root, value_interface, optional_interface, TestListInstanceFactory(), 0)
        self.assertEqual(value_interface.val, 42)

    def test_add_controller(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()

        controller = Controller.ListController('test', root, value_interface, None, TestListInstanceFactory(), 0)
        self.assertEqual(len(controller._controllers), 0)
        controller.add_controller('sub', controller)
        self.assertEqual(len(controller._controllers), 1)

    def test_model_to_view(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()
        list_instance_factory = TestListInstanceFactory()

        controller = Controller.ListController('test', root, value_interface, None, list_instance_factory, 0)

        with self.assertRaises(AssertionError):
            controller.model_to_view({})

        controller = Controller.ListController('test', root, value_interface, optional_interface, list_instance_factory, 0)

        controller.model_to_view({})
        self.assertFalse(optional_interface.val)

        controller.model_to_view({'test': []})
        self.assertEqual(value_interface.val, '0')
        self.assertEqual(len(list_instance_factory.instances), 0)
        self.assertTrue(optional_interface.val)

        controller.model_to_view({'test': ['1', '2', '3']})
        self.assertEqual(value_interface.val, '3')
        self.assertEqual(len(list_instance_factory.instances), 3)
        self.assertTrue(optional_interface.val)
        self.assertEqual(list_instance_factory.values[0].val, '1')
        self.assertEqual(list_instance_factory.values[1].val, '2')
        self.assertEqual(list_instance_factory.values[2].val, '3')

    def test_view_to_model(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()
        list_instance_factory = TestListInstanceFactory()

        controller = Controller.ListController('test', root, value_interface, optional_interface, list_instance_factory, 0)

        optional_interface.val = False
        self.assertIsNone(controller.view_to_model())

        optional_interface.val = True
        self.assertEqual(controller.view_to_model(), [])

        value_interface.val = 2
        controller.event_handler()
        list_instance_factory.values[0].val = 1
        list_instance_factory.values[1].val = 2
        self.assertEqual(controller.view_to_model(), [1, 2])

    def test_event_handlers(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()
        list_instance_factory = TestListInstanceFactory()

        controller = Controller.ListController('test', root, value_interface, optional_interface, list_instance_factory, 0)
        self.assertEqual(len(list_instance_factory.instances), 0)

        value_interface.val = 2
        controller.event_handler()
        self.assertEqual(len(list_instance_factory.instances), 2)

        optional_interface.val = False
        controller.optional_handler()
        self.assertEqual(len(list_instance_factory.instances), 0)
        self.assertEqual(value_interface.val, 2)

        optional_interface.val = True
        controller.optional_handler()
        self.assertEqual(len(list_instance_factory.instances), 2)
        self.assertEqual(value_interface.val, 2)



import typing
from typing import Optional, Callable, List, Tuple

import wx
import wx.adv

from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.view.AbstractView import AbstractView

ControlList = typing.Dict[str, typing.Union[wx.TextCtrl, wx.CheckBox, wx.StaticBitmap, wx.ComboBox, wx.StaticText, wx.SpinCtrl, List[Tuple[int, wx.CheckBox]],
                                            str, wx.adv.DatePickerCtrl, wx.adv.TimePickerCtrl]]


class WxPythonView(AbstractView, OptionalInterface):
    structure_changed: Callable = None

    def __init__(self, name: str, controls: ControlList, container=False):
        self._name = name
        self._controls = controls
        self._container = container

    def register_optional_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self._controls['optional'].GetTopLevelParent().Freeze()
            callback()
            self.enable(self.get_has_value())
            self._controls['optional'].GetTopLevelParent().Thaw()

        if 'optional' in self._controls:
            self._controls.get('optional').Bind(wx.EVT_CHECKBOX, event_closure)

    def register_change_event(self, callback: Callable):
        pass

    def get_has_value(self) -> bool:
        if 'optional' in self._controls:
            return self._controls.get('optional').GetValue()
        else:
            return True

    def set_has_value(self, val: bool):
        if 'optional' in self._controls:
            self._controls.get('optional').SetValue(val)
            self.enable(val)

    def realize(self) -> 'WxPythonView':
        return self

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        raise NotImplementedError()

    def enable(self, enabled: bool):
        return

    def destroy(self):
        for name, control in self._controls.items():
            if name == 'optional':
                continue
            if isinstance(control, wx.Object):
                control.Destroy()

    def get_name(self) -> str:
        return self._name

    def set_visible(self, visible, recursive=True):
        for control in self._controls.values():
            if isinstance(control, wx.Window):
                control.Show(visible)

    def _create_sizer(self, orientation: int = wx.HORIZONTAL) -> wx.BoxSizer:
        sizer = wx.BoxSizer(orientation)
        if orientation == wx.HORIZONTAL:
            if 'icon' in self._controls:
                sizer.Add(self._controls['icon'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=2)
            sizer.Add(self._controls['name'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
        else:
            if 'icon' in self._controls:
                sub_sizer = wx.BoxSizer(wx.HORIZONTAL)
                sub_sizer.Add(self._controls['icon'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=2)
                sub_sizer.Add(self._controls['name'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
                sizer.Add(sub_sizer)
            else:
                sizer.Add(self._controls['name'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)

        return sizer


class WxPythonValueView(WxPythonView, ValueInterface):
    def __init__(self, name: str, controls: ControlList):
        super(WxPythonValueView, self).__init__(name, controls)

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        sizer = self._create_sizer()
        sizer.Add(self._controls['value'], proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        if self._controls.get('style') == 'hidden':
            sizer.ShowItems(False)
        return sizer

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self.__update_tooltip()
            callback()

        self._controls['value'].Bind(wx.EVT_TEXT, event_closure)

    def get_value(self) -> str:
        return self._controls['value'].GetValue()

    def set_value(self, val: str):
        self._controls['value'].SetValue(val)
        self.__update_tooltip()

    def enable(self, enabled: bool):
        self._controls['value'].Enable(enabled)

    def __update_tooltip(self):
        if isinstance(self.get_value(), str) and len(self.get_value()) > 10:
            tool_tip = self._controls['value'].GetToolTip()
            if tool_tip is not None:
                previous_tooltip = self._controls['value'].GetToolTip().GetTip().split('\n')
                previous_tooltip = previous_tooltip[-1]
            else:
                previous_tooltip = ''
            self._controls['value'].SetToolTip('\n'.join([self.get_value(), previous_tooltip]))


class WxPythonValueSelectionView(WxPythonValueView):
    def __init__(self, name: str, controls: ControlList):
        super(WxPythonValueSelectionView, self).__init__(name, controls)

    def get_value(self) -> str:
        return self._controls['value'].GetStringSelection()

    def set_value(self, val: str):
        self._controls['value'].SetStringSelection(val)


class WxPythonHexStringView(WxPythonView, ValueInterface):
    CHARS_PER_HEX_DIGIT = 3

    def __init__(self, name: str, controls: ControlList, minimum: Optional[int], maximum: Optional[int]):
        super(WxPythonHexStringView, self).__init__(name, controls)

        self._real_value = b''
        self._controls['selector'].Bind(wx.EVT_CHECKBOX, self.hex_selector_changed)
        self._hex = self._is_hex()
        self._minimum = minimum
        if self._minimum is None:
            self._minimum = 0
        self._maximum = maximum
        self._update_length()

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        sizer = self._create_sizer()
        sizer.Add(self._controls['selector'], flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
        sizer.Add(self._controls['value'], proportion=1, flag=wx.ALL | wx.EXPAND, border=5)
        return sizer

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self.text_changed()
            callback()

        self._controls['value'].Bind(wx.EVT_TEXT, event_closure)

    # noinspection PyUnusedLocal
    def hex_selector_changed(self, e: wx.CommandEvent):
        self._hex_selector_changed()

    def text_changed(self):
        if self._hex:
            try:
                self._real_value = bytes.fromhex(self._controls['value'].GetValue())
            except ValueError:
                pass
        else:
            self._real_value = self._controls['value'].GetValue().encode('latin-1')

    def _update_control(self):
        if self._hex:
            val = self._real_value.hex()
            val = ' '.join(val[i:i + 2] for i in range(0, len(val), 2))
        else:
            val = self._real_value.decode('latin-1', errors='replace')

        self._controls['value'].ChangeValue(val)

        self.__update_tooltip()

    def _hex_selector_changed(self):
        if self._hex != self._is_hex():
            if not self._hex:
                # Was ASCII before, now is hex
                if self._maximum:
                    self._maximum *= self.CHARS_PER_HEX_DIGIT
                self._minimum *= self.CHARS_PER_HEX_DIGIT
            else:
                if self._maximum:
                    self._maximum //= self.CHARS_PER_HEX_DIGIT
                self._minimum //= self.CHARS_PER_HEX_DIGIT

            self._update_length()
            self._hex = self._is_hex()
            self._update_control()

    def __update_tooltip(self):
        if len(self.get_value()) > 10:
            tool_tip = self._controls['value'].GetToolTip()
            if tool_tip is not None:
                previous_tooltip = self._controls['value'].GetToolTip().GetTip().split('\n')
                previous_tooltip = previous_tooltip[-1]
            else:
                previous_tooltip = ''
            self._controls['value'].SetToolTip('\n'.join([str(self.get_value()), previous_tooltip]))

    def get_value(self) -> bytes:
        return self._real_value

    def set_value(self, val: bytes):
        assert isinstance(val, bytes)
        self._real_value = val
        self._update_control()

    def enable(self, enabled: bool):
        self._controls['value'].Enable(enabled)

    def _is_hex(self) -> bool:
        return self._controls['selector'].GetValue()

    def _update_length(self):
        if self._maximum:
            self._controls['value'].SetToolTip(f"Minimum characters: {self._minimum}, maximum characters: {self._maximum}")
            self._controls['value'].SetMaxLength(self._maximum)


class WxPythonBooleanView(WxPythonView, ValueInterface):
    def __init__(self, name: str, controls: ControlList):
        super(WxPythonBooleanView, self).__init__(name, controls)

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()

        self._controls['value'].Bind(wx.EVT_CHECKBOX, event_closure)

    def get_value(self) -> bool:
        return self._controls['value'].GetValue()

    def set_value(self, val: bool):
        self._controls['value'].SetValue(val)

    def enable(self, enabled: bool):
        self._controls['value'].Enable(enabled)

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        sizer = self._create_sizer()
        sizer.Add(self._controls['value'], proportion=1, flag=wx.ALL | wx.EXPAND, border=5)
        return sizer


class WxPythonBitstringView(WxPythonView, BitstringInterface):
    def __init__(self, name: str, controls: ControlList, parent: wx.Window):
        super(WxPythonBitstringView, self).__init__(name, controls)
        self._parent = parent

    def enable(self, enabled: bool):
        for _, checkbox in self._controls['checkboxes']:
            checkbox.Enable(enabled)

    def get_values(self) -> typing.List[int]:
        values = []
        for bit, checkbox in self._controls['checkboxes']:
            if checkbox.GetValue():
                values.append(bit)
        return values

    def set_values(self, values: typing.List[int]):
        for bit, checkbox in self._controls['checkboxes']:
            checkbox.SetValue(bit in values)

    def get_sizer(self, recursive: bool) -> wx.Sizer:
        sizer = self._create_sizer()
        if self._controls['name'].IsShown():
            bits_sizer = wx.StaticBoxSizer(wx.VERTICAL, self._parent, "Bits")
        else:
            bits_sizer = wx.BoxSizer(wx.VERTICAL)

        for _, checkbox in self._controls['checkboxes']:
            bits_sizer.Add(checkbox)
        sizer.Add(bits_sizer)

        return sizer

    def destroy(self):
        super(WxPythonBitstringView, self).destroy()
        for _, checkbox in self._controls['checkboxes']:
            checkbox.Destroy()

    def set_visible(self, visible, recursive=True):
        super(WxPythonBitstringView, self).set_visible(visible, recursive)
        for _, checkbox in self._controls['checkboxes']:
            checkbox.Show(visible)



import re

from setuptools import find_packages
from setuptools import setup


def find_version():
    return re.search(r'^__version__ = "(.*)"$',
                     open('asn1editor/__init__.py', 'r').read(),
                     re.MULTILINE).group(1)


setup(name='asn1editor',
      version=find_version(),
      description='ASN.1 editor framework with support for encoding and decoding various codecs.',
      long_description=open('README.md', 'r').read(),
      long_description_content_type='text/markdown',
      author='Florian Fetz',
      author_email='florian.fetz@gmail.com',
      license='MIT',
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development'
      ],
      keywords=['ASN.1', 'asn1', 'editor'],
      url='https://github.com/futsch1/asn1editor',
      packages=find_packages(exclude=['tests', 'example']),
      include_package_data=True,
      install_requires=[
          'asn1tools>=0.153.0',
          'wxpython>=4.1.0'
      ],
      test_suite="tests",
      python_requires='>=3.6',
      entry_points={
          'console_scripts': ['asn1editor=asn1editor.__init__:_wx_python_editor']
      })



from unittest import TestCase

from asn1editor.controller import Controller
from asn1editor.controller import Converter
from tests.controller.test_listController import TestListInstanceFactory
from tests.controller.test_valueBoolControllers import TestValueInterface, TestOptionalInterface


class TestListInstanceFactoryContainer:
    def __init__(self):
        self.values = {}
        self.instances = {}

    def create(self, i, p: Controller.ListController):
        self.values[i] = TestValueInterface()
        self.instances[i] = Controller.ContainerController(str(i), p, None)

    def destroy(self, i):
        del self.values[i]
        del self.instances[i]


class TestContainerController(TestCase):
    def test_add_controller(self):
        root = Controller.RootController('root')

        controller = Controller.ContainerController('test', root, None)
        self.assertEqual(len(controller._controllers), 0)
        controller.add_controller('sub', controller)
        self.assertEqual(len(controller._controllers), 1)

    def test_model_to_view_value(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface_container = TestOptionalInterface()
        optional_interface_value = TestOptionalInterface()

        container = Controller.ContainerController('test', root, None)
        Controller.ValueController('test', container, value_interface, None, Converter.Str(0, 'default'))
        container.model_to_view({'test': {'test': 'test'}})
        self.assertEqual(value_interface.val, 'test')

        container = Controller.ContainerController('test', root, optional_interface_container)
        Controller.ValueController('test', container, value_interface, optional_interface_value, Converter.Str(0, 'default'))
        container.model_to_view({'test': {'test': 'test'}})
        self.assertTrue(optional_interface_container.val)
        self.assertTrue(optional_interface_value.val)

        container.model_to_view({'test': {}})
        self.assertTrue(optional_interface_container.val)
        self.assertFalse(optional_interface_value.val)

        container.model_to_view({})
        self.assertFalse(optional_interface_container.val)

        container = Controller.ContainerController('test', root, None)
        container2 = Controller.ContainerController('test2', container, None)
        Controller.ValueController('test', container2, value_interface, None, Converter.Str(0, 'default'))
        container.model_to_view({'test': {'test2': {'test': 'test'}}})
        self.assertEqual(value_interface.val, 'test')

    def test_model_to_view_list(self):
        root = Controller.RootController('root')
        list_value_interface = TestValueInterface()
        value_interface = TestValueInterface()
        list_instance_factory = TestListInstanceFactory()

        container = Controller.ContainerController('test', root, None)
        Controller.ListController('test_list', container, list_value_interface, None, list_instance_factory, 0)

        container.model_to_view({'test': {'test_list': []}})
        self.assertEqual(list_value_interface.val, '0')
        self.assertEqual(len(list_instance_factory.instances), 0)

        container.model_to_view({'test': {'test_list': [1]}})
        self.assertEqual(list_value_interface.val, '1')
        self.assertEqual(len(list_instance_factory.instances), 1)
        self.assertEqual(list_instance_factory.values[0].val, 1)

        Controller.ValueController('test_val', container, value_interface, None, Converter.Str(0, 'default'))
        container.model_to_view({'test': {'test_list': [1], 'test_val': 'test'}})
        self.assertEqual(value_interface.val, 'test')

        list_instance_factory = TestListInstanceFactoryContainer()
        list_value_interface.val = 0
        container = Controller.ListController('test_list', root, list_value_interface, None, list_instance_factory, 0)

        container.model_to_view({'test_list': []})
        self.assertEqual(list_value_interface.val, '0')
        self.assertEqual(len(list_instance_factory.instances), 0)

        container.model_to_view({'test_list': [{}]})
        self.assertEqual(list_value_interface.val, '1')
        self.assertEqual(len(list_instance_factory.instances), 1)

    def test_view_to_model_value(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface_container = TestOptionalInterface()
        optional_interface_value = TestOptionalInterface()

        container = Controller.ContainerController('test', root, None)
        Controller.ValueController('test', container, value_interface, None, Converter.Str(0, 'default'))
        value_interface.val = 'test'

        self.assertEqual(container.view_to_model(), {'test': 'test'})

        container = Controller.ContainerController('test', root, optional_interface_container)
        Controller.ValueController('test', container, value_interface, optional_interface_value, Converter.Str(0, 'default'))

        optional_interface_value.val = True
        optional_interface_container.val = True
        self.assertEqual(container.view_to_model(), {'test': 'default'})

        optional_interface_value.val = False
        optional_interface_container.val = False
        self.assertIsNone(container.view_to_model())

        optional_interface_value.val = False
        optional_interface_container.val = True
        self.assertEqual(container.view_to_model(), {})

        container = Controller.ContainerController('test', root, None)
        container2 = Controller.ContainerController('test2', container, None)
        Controller.ValueController('test', container2, value_interface, None, Converter.Str(0, 'default'))

        value_interface.val = 'test'
        self.assertEqual(container.view_to_model(), {'test2': {'test': 'test'}})

    def test_view_to_model_list(self):
        root = Controller.RootController('root')
        list_value_interface = TestValueInterface()
        value_interface = TestValueInterface()
        list_instance_factory = TestListInstanceFactory()

        container = Controller.ContainerController('test', root, None)
        list_controller = Controller.ListController('test_list', container, list_value_interface, None, list_instance_factory, 0)

        self.assertEqual(container.view_to_model(), {'test_list': []})

        list_value_interface.val = 1
        list_controller.event_handler()
        list_instance_factory.values[0].val = 1

        self.assertEqual(container.view_to_model(), {'test_list': [1]})

        container.model_to_view({'test': {'test_list': [1]}})
        self.assertEqual(list_value_interface.val, '1')
        self.assertEqual(len(list_instance_factory.instances), 1)
        self.assertEqual(list_instance_factory.values[0].val, 1)

        Controller.ValueController('test_val', container, value_interface, None, Converter.Str(0, 'default'))
        value_interface.val = 'test'
        self.assertEqual(container.view_to_model(), {'test_list': [1], 'test_val': 'test'})



import typing
from datetime import date, time


class Converter:
    def __init__(self, minimum: typing.Optional[typing.Union[int, float]], default: typing.Optional[typing.Union[str, bytes, int, float, date, time]]):
        self._default = default
        self._minimum = minimum
        if self._minimum is None:
            self._minimum = 0

    @staticmethod
    def to_view(val: typing.Any) -> typing.Union[str, bytes]:  # pragma: no cover
        raise NotImplementedError

    @staticmethod
    def from_view(val: typing.Union[str, bytes]) -> typing.Any:  # pragma: no cover
        raise NotImplementedError

    @staticmethod
    def default() -> typing.Union[str, bytes, int, float]:  # pragma: no cover
        raise NotImplementedError


class Str(Converter):
    @staticmethod
    def to_view(val: str) -> str:
        return val

    @staticmethod
    def from_view(val: str) -> str:
        return val

    def default(self) -> str:
        return ' ' * self._minimum if not self._default else self._default


class Any(Converter):
    @staticmethod
    def to_view(val: typing.Any) -> typing.Any:
        return val

    @staticmethod
    def from_view(val: typing.Any) -> typing.Any:
        return val

    def default(self) -> typing.Any:
        return self._default


class Int(Converter):
    @staticmethod
    def to_view(val: int) -> str:
        return str(val)

    @staticmethod
    def from_view(val: str) -> int:
        try:
            i = int(val)
        except ValueError:
            i = 0
        return i

    def default(self) -> int:
        return self._minimum if not self._default else self.from_view(self._default)


class Float(Converter):
    @staticmethod
    def to_view(val: float) -> str:
        return str(val)

    @staticmethod
    def from_view(val: str) -> float:
        try:
            i = float(val)
        except ValueError:
            i = 0
        return i

    def default(self) -> float:
        return self._minimum if not self._default else self._default


class ByteString(Converter):
    @staticmethod
    def to_view(val: bytes) -> bytes:
        return val

    @staticmethod
    def from_view(val: bytes) -> bytes:
        return val

    def default(self) -> bytes:
        if isinstance(self._default, str):
            self._default = self._default.encode('latin-1')
        return b' ' * self._minimum if not self._default else self._default
